import re
from abc import ABC, abstractmethod

from .. import emoji_data
from ..models import CANCEL, DEFAULT, Action, CharacterEntry, Shortcut


class Selector(ABC):
    @staticmethod
    def best_option(name: str | None = None) -> "Selector":
        from .bemenu import Bemenu
        from .choose import Choose
        from .dmenu import DMenu
        from .fuzzel import Fuzzel
        from .hyprlauncher import Hyprlauncher
        from .rofi import Rofi
        from .tofi import Tofi
        from .wmenu import WMenu
        from .wofi import Wofi

        available_selectors = [Rofi, Wofi, Fuzzel, Bemenu, Tofi, Hyprlauncher, DMenu, WMenu, Choose]

        if name is not None:
            try:
                return next(selector for selector in available_selectors if selector.name() == name)()
            except StopIteration:
                raise NoSelectorFoundException()
        else:
            try:
                return next(selector for selector in available_selectors if selector.supported())()
            except StopIteration:
                raise NoSelectorFoundException()

    @staticmethod
    @abstractmethod
    def supported() -> bool:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @abstractmethod
    def show_character_selection(
        self,
        characters: list[CharacterEntry],
        recent_characters: list[str],
        prompt: str,
        show_description: bool,
        use_icons: bool,
        keybindings: dict[Action, str],
        additional_args: list[str],
    ) -> tuple[Action | DEFAULT | CANCEL, list[str] | Shortcut]:
        pass

    @abstractmethod
    def show_skin_tone_selection(
        self,
        selected_emoji: str,
        prompt: str,
        show_description: bool,
        use_icons: bool,
        additional_args: list[str],
    ) -> tuple[int, str]:
        pass

    @abstractmethod
    def show_action_menu(self, additional_args: list[str]) -> list[Action]:
        pass

    @staticmethod
    def _basic_format_skin_tones(selected_emoji: str) -> list[str]:
        return [
            f"{selected_emoji}{modifier} {emoji_data.fitzpatrick_modifiers[modifier]}"
            for modifier in emoji_data.fitzpatrick_modifiers
        ]

    def basic_format_characters(self, characters: list[CharacterEntry], strip_tags: bool = True) -> list[str]:
        return [
            f"{entry.character} {entry.description.replace('<small>', '').replace('</small>', '') if strip_tags else entry.description}"
            for entry in characters
        ]

    def _extract_char_from_output(self, line: str) -> str:
        return re.match(r"^(?:\u200e(?! ))?(?P<char>.[^ ]*)( .*|$)", line).group("char")


class NoSelectorFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to show the selection. Please check the required dependencies."



from dataclasses import dataclass
from enum import Enum


class Action(Enum):
    TYPE = "type"
    COPY = "copy"
    CLIPBOARD = "clipboard"
    TYPE_NUMERICAL = "type-numerical"
    UNICODE = "unicode"
    COPY_UNICODE = "copy-unicode"
    STDOUT = "print"
    MENU = "menu"

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class CANCEL:
    def __eq__(self, other):
        return isinstance(other, CANCEL)


class DEFAULT:
    def __eq__(self, other):
        return isinstance(other, DEFAULT)


@dataclass(slots=True)
class Shortcut:
    index: int


@dataclass(slots=True)
class CharacterEntry:
    character: str
    description: str | None = None

    def merge(self, other: "CharacterEntry"):
        if self == other:
            return self

        if self.character != other.character:
            raise Exception("Cannot merge different characters")

        if other.description:
            if self.description:
                self.description = f"{self.description}, {other.description}"
            else:
                self.description = other.description



from subprocess import run

from ..abstractionhelper import is_installed, is_wayland
from ..typer.typer import Typer
from .clipboarder import Clipboarder


class XSelClipboarder(Clipboarder):
    @staticmethod
    def supported() -> bool:
        return not is_wayland() and is_installed("xsel")

    @staticmethod
    def name() -> str:
        return "xsel"

    def copy_characters_to_clipboard(self, characters: str) -> None:
        run(["xsel", "-i", "-b"], input=characters, encoding="utf-8")

    def copy_paste_characters(self, characters: str, active_window: str, typer: Typer) -> None:
        old_clipboard_content = run(args=["xsel", "-o", "-b"], capture_output=True, encoding="utf-8")
        old_primary_content = run(args=["xsel", "-o", "-p"], capture_output=True, encoding="utf-8")

        run(args=["xsel", "-i", "-b"], input=characters, encoding="utf-8")
        run(args=["xsel", "-i", "-p"], input=characters, encoding="utf-8")

        typer.insert_from_clipboard(active_window)

        if old_clipboard_content.returncode == 0:
            run(args=["xsel", "-i", "-b"], input=old_clipboard_content.stdout, encoding="utf-8")
        if old_primary_content.returncode == 0:
            run(args=["xsel", "-i", "-p"], input=old_primary_content.stdout, encoding="utf-8")



import os

from .mode import ModeRofimoji
from .standalone import StandaloneRofimoji


def main():
    if "ROFI_RETV" not in os.environ:
        StandaloneRofimoji().standalone()
    else:
        ModeRofimoji().mode()


if __name__ == "__main__":
    main()



import os
import platform
import shutil


def is_installed(executable: str) -> bool:
    return shutil.which(executable) is not None


def is_macos() -> bool:
    return platform.system() == "Darwin"


def is_wayland() -> bool:
    return "WAYLAND_DISPLAY" in os.environ



from subprocess import run

from ..abstractionhelper import is_installed
from ..emoji_data import fitzpatrick_modifiers
from ..models import CANCEL, DEFAULT, Action, CharacterEntry, Shortcut
from .selector import Selector


class Rofi(Selector):
    @staticmethod
    def supported() -> bool:
        return is_installed("rofi")

    @staticmethod
    def name() -> str:
        return "rofi"

    def show_character_selection(
        self,
        characters: list[CharacterEntry],
        recent_characters: list[str],
        prompt: str,
        show_description: bool,
        use_icons: bool,
        keybindings: dict[Action, str],
        additional_args: list[str],
    ) -> tuple[Action | DEFAULT | CANCEL, list[str] | Shortcut]:
        parameters = [
            "rofi",
            "-dmenu",
            "-markup-rows",
            "-i",
            "-multi-select",
            "-no-custom",
            "-ballot-unselected-str",
            "",
            "-format",
            "i",
            "-p",
            prompt,
            "-kb-custom-11",
            keybindings[Action.COPY],
            "-kb-custom-12",
            keybindings[Action.TYPE],
            "-kb-custom-13",
            keybindings[Action.CLIPBOARD],
            "-kb-custom-14",
            keybindings[Action.TYPE_NUMERICAL],
            "-kb-custom-15",
            keybindings[Action.UNICODE],
            "-kb-custom-16",
            keybindings[Action.COPY_UNICODE],
            *additional_args,
        ]

        if recent_characters:
            parameters.extend(["-mesg", self.__format_recent_characters(recent_characters)])

        rofi = run(
            parameters,
            input="\n".join(self.__format_characters(characters, use_icons, show_description)),
            capture_output=True,
            encoding="utf-8",
        )

        if 10 <= rofi.returncode <= 19:
            return DEFAULT(), Shortcut(rofi.returncode - 10)

        action: Action | DEFAULT | CANCEL
        match rofi.returncode:
            case 1:
                action = CANCEL()
            case 20:
                action = Action.COPY
            case 21:
                action = Action.TYPE
            case 22:
                action = Action.CLIPBOARD
            case 23:
                action = Action.UNICODE
            case 24:
                action = Action.COPY_UNICODE
            case 25:
                action = Action.TYPE_NUMERICAL
            case _:
                action = DEFAULT()

        return action, [characters[int(index)].character for index in rofi.stdout.splitlines()]

    def __format_characters(
        self, characters: list[CharacterEntry], use_icons: bool, show_description: bool
    ) -> list[str]:
        if use_icons and not show_description:
            return [f"\0meta\x1f{entry.description}\x1ficon\x1f<span>{entry.character}</span>" for entry in characters]
        elif use_icons and show_description:
            return [f"{entry.description}\0icon\x1f<span>{entry.character}</span>" for entry in characters]
        elif not use_icons and show_description:
            return self.basic_format_characters(characters, strip_tags=False)
        else:
            return [f"{entry.character}\0meta\x1f{entry.description}" for entry in characters]

    def __format_recent_characters(self, recent_characters: list[str]) -> str:
        pairings = [f"\u200e{(index + 1) % 10}: {character}" for index, character in enumerate(recent_characters)]

        return " | ".join(pairings)

    def show_skin_tone_selection(
        self,
        selected_emoji: str,
        prompt: str,
        show_description: bool,
        use_icons: bool,
        additional_args: list[str],
    ) -> tuple[int, str]:
        formatted = self.__format_skin_tones(
            selected_emoji,
            show_description,
            use_icons,
        )
        rofi = run(
            ["rofi", "-dmenu", "-markup-rows", "-i", "-no-custom", "-format", "i", "-p", prompt, *additional_args],
            input="\n".join(formatted),
            capture_output=True,
            encoding="utf-8",
        )

        if rofi.returncode == 1:
            return rofi.returncode, ""

        return rofi.returncode, (selected_emoji + list(fitzpatrick_modifiers.keys())[int(rofi.stdout.strip())])

    def __format_skin_tones(self, selected_emoji: str, show_description: bool, use_icons: bool) -> list[str]:
        if use_icons and not show_description:
            return [
                f"\0meta\x1f{description}\x1ficon\x1f<span>{selected_emoji}{modifier}</span>"
                for (modifier, description) in fitzpatrick_modifiers.items()
            ]
        elif use_icons and show_description:
            return [
                f"{description}\0icon\x1f<span>{selected_emoji}{modifier}</span>"
                for (modifier, description) in fitzpatrick_modifiers.items()
            ]
        elif not use_icons and show_description:
            return [
                f"{selected_emoji}{modifier} {description}" for (modifier, description) in fitzpatrick_modifiers.items()
            ]
        else:
            return [
                f"{selected_emoji}{modifier}\0meta\x1f{description}"
                for (modifier, description) in fitzpatrick_modifiers.items()
            ]

    def show_action_menu(self, additional_args: list[str]) -> list[Action]:
        rofi = run(
            [
                "rofi",
                "-dmenu",
                "-multi-select",
                "-no-custom",
                "-ballot-unselected-str",
                "",
                "-i",
                *additional_args,
            ],
            input="\n".join([str(it) for it in Action if it != Action.MENU]),
            capture_output=True,
            encoding="utf-8",
        )

        return [Action(action) for action in rofi.stdout.strip().split("\n")]



import os
from pathlib import Path

config_home = Path(os.environ.get("XDG_CONFIG_HOME", "~/.config")).expanduser()
config_global = [Path(dir) for dir in os.environ.get("XDG_CONFIG_DIRS", "/etc/xdg").split(":")]
data_home = Path(os.environ.get("XDG_DATA_HOME", "~/.local/share")).expanduser()
cache_home = Path(os.environ.get("XDG_CACHE_HOME", "~/.cache")).expanduser()

config_file_locations = [str(directory / "rofimoji.rc") for directory in [config_home] + config_global]
recents_file_location = data_home / "rofimoji" / "recent"
frecency_file_location = data_home / "rofimoji" / "frecency"
custom_additional_files_location = data_home / "rofimoji" / "data"

cache_file_location = cache_home / "rofimoji"



from ..typer.typer import Typer
from .clipboarder import Clipboarder


class NoopClipboarder(Clipboarder):
    @staticmethod
    def supported() -> bool:
        return True

    @staticmethod
    def name() -> str:
        return "noop"

    def copy_characters_to_clipboard(self, characters: str) -> None:
        raise NoClipboarderFoundException()

    def copy_paste_characters(self, characters: str, active_window: str, typer: Typer) -> None:
        raise NoClipboarderFoundException()


class NoClipboarderFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to copy to clipboard. Please check the required dependencies."



from subprocess import run

from ..abstractionhelper import is_installed, is_wayland
from .typer import Typer


class XDoToolTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return not is_wayland() and is_installed("xdotool")

    @staticmethod
    def name() -> str:
        return "xdotool"

    def get_active_window(self) -> str:
        return run(args=["xdotool", "getactivewindow"], capture_output=True, encoding="utf-8").stdout[:-1]

    def type_characters(self, characters: str, active_window: str) -> None:
        # There is a timing issue for some applications that do not correctly handle inputs this fast.
        # Slowing down xdotool from 12ms per keystroke to 50ms per keystroke resolves the issue.
        run(["xdotool", "windowactivate", active_window, "type", "--delay=50ms", "--clearmodifiers", characters])

    def insert_from_clipboard(self, active_window: str) -> None:
        run(
            [
                "xdotool",
                "windowfocus",
                "--sync",
                active_window,
                "key",
                "--clearmodifiers",
                "Shift+Insert",
                "sleep",
                "0.05",
            ]
        )

    def type_numerical(self, codepoints: list[int], active_window: str) -> None:
        codepoint_list = " ".join([f"U{codepoint:x}" for codepoint in codepoints])

        run(
            [
                "xdotool",
                "windowfocus",
                "--sync",
                active_window,
                "key",
                "--clearmodifiers",
                "Ctrl+Shift",
                codepoint_list,
                "sleep",
                "0.05",
            ]
        )



from subprocess import run

from ..abstractionhelper import is_installed, is_wayland
from .typer import Typer


class WlImeTypeTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("wl-ime-type")

    @staticmethod
    def name() -> str:
        return "wl-ime-type"

    def get_active_window(self) -> str:
        return "not possible with wl-ime-type"

    def type_characters(self, characters: str, active_window: str) -> None:
        run(["wl-ime-type", characters])

    def insert_from_clipboard(self, active_window: str) -> None:
        return "not possible with wl-ime-type"

    def type_numerical(self, codepoints: list[int], active_window: str) -> None:
        return "not possible with wl-ime-type"



import argparse
import shlex

import configargparse

from . import __version__
from .models import Action
from .paths import config_file_locations


def parse_arguments_strict() -> argparse.Namespace:
    return __parse_arguments(only_known=True)


def parse_arguments_flexible() -> argparse.Namespace:
    return __parse_arguments(only_known=False)


def __parse_arguments(only_known: bool) -> argparse.Namespace:
    parser = configargparse.ArgumentParser(
        description="Select, insert or copy Unicode characters using rofi.",
        default_config_files=config_file_locations,
    )
    parser.add_argument("--version", action="version", version="rofimoji " + __version__)
    parser.add_argument(
        "--action",
        "-a",
        dest="actions",
        action="store",
        type=Action,
        choices=list(Action),
        default=[Action.TYPE],
        nargs="*",
        metavar="ACTION",
        help="How to insert the chosen characters. More than one action may be specified in "
        "a space separated list (for example: `--action type copy`). Options: "
        + ", ".join(f'"{a.value}"' for a in Action),
    )
    parser.add_argument(
        "--skin-tone",
        "-s",
        dest="skin_tone",
        action="store",
        choices=["neutral", "light", "medium-light", "moderate", "dark brown", "black", "ask"],
        default="ask",
        help='Decide on a skin-tone for all supported emojis. If not set (or set to "ask"), '
        "you will be asked for each one ",
    )
    parser.add_argument(
        "--files",
        "-f",
        dest="files",
        action="store",
        default=[
            "emojis_smileys_emotion",
            "emojis_people_body",
            "emojis_animals_nature",
            "emojis_food_drink",
            "emojis_activities",
            "emojis_travel_places",
            "emojis_objects",
            "emojis_symbols",
            "emojis_flags",
        ],
        nargs="+",
        metavar="FILE",
        help="Read characters from this file instead, one entry per line",
    )
    parser.add_argument("--prompt", "-r", dest="prompt", action="store", default="😀 ", help="Set rofimoji's  prompt")
    parser.add_argument(
        "--selector-args",
        dest="selector_args",
        action="store",
        default=False,
        help="A string of arguments to give to the selector",
    )
    parser.add_argument(
        "--max-recent",
        dest="max_recent",
        action="store",
        type=int,
        default=10,
        help="Show at most this number of recently used characters (cannot be larger than 10)",
    )
    parser.add_argument(
        "--no-frecency", dest="frecency", action="store_false", help="Don't show frequently used characters first"
    )
    parser.set_defaults(frecency=True)
    parser.add_argument(
        "--only-official",
        dest="use_additional",
        action="store_false",
        help="Use only the official Unicode descriptions",
    )
    parser.set_defaults(use_additional=True)
    parser.add_argument(
        "--hidden-descriptions",
        dest="show_description",
        action="store_false",
        help="Show only the character without its description",
    )
    parser.set_defaults(show_description=True)
    parser.add_argument(
        "--use-icons",
        dest="use_icons",
        action="store_true",
        help="Use rofi's icon to show the character",
    )
    parser.set_defaults(use_icons=False)
    parser.add_argument(
        "--selector",
        dest="selector",
        action="store",
        type=str,
        choices=["rofi", "wofi", "fuzzel", "dmenu", "tofi", "bemenu", "wmenu", "hyprlauncher", "choose"],
        default=None,
        help="Choose the application to select the characters with",
    )
    parser.add_argument(
        "--clipboarder",
        dest="clipboarder",
        action="store",
        type=str,
        choices=["xsel", "xclip", "wl-copy", "pbcopy"],
        default=None,
        help="Choose the application to access the clipboard with",
    )
    parser.add_argument(
        "--typer",
        dest="typer",
        action="store",
        type=str,
        choices=["xdotool", "wtype", "ydotool", "cliclick", "wl-ime-type"],
        default=None,
        help="Choose the application to type with",
    )
    parser.add_argument(
        "--keybinding-copy",
        dest="keybinding_copy",
        action="store",
        type=str,
        default="Alt+c",
        help="Choose the keyboard shortcut to copy the character to the clipboard",
    )
    parser.add_argument(
        "--keybinding-type",
        dest="keybinding_type",
        action="store",
        type=str,
        default="Alt+t",
        help="Choose the keyboard shortcut to directly type the character",
    )
    parser.add_argument(
        "--keybinding-clipboard",
        dest="keybinding_clipboard",
        action="store",
        type=str,
        default="Alt+p",
        help="Choose the keyboard shortcut to insert the character via the clipboard",
    )
    parser.add_argument(
        "--keybinding-type-numerical",
        dest="keybinding_type_numerical",
        action="store",
        type=str,
        default="Alt+n",
        help="Choose the keyboard shortcut to directly type the character's unicode codepoint numerically "
        "with Ctrl+Shift+U<codepoint>",
    )
    parser.add_argument(
        "--keybinding-unicode",
        dest="keybinding_unicode",
        action="store",
        type=str,
        default="Alt+u",
        help="Choose the keyboard shortcut to directly type the character's unicode codepoint",
    )
    parser.add_argument(
        "--keybinding-copy-unicode",
        dest="keybinding_copy_unicode",
        action="store",
        type=str,
        default="Alt+i",
        help="Choose the keyboard shortcut to copy the character's unicode codepoint to the clipboard",
    )

    if only_known:
        parsed_args = parser.parse_args()
    else:
        parsed_args, _ = parser.parse_known_args()

    if parsed_args.selector_args:
        parsed_args.selector_args = shlex.split(parsed_args.selector_args)
    else:
        parsed_args.selector_args = []

    parsed_args.keybindings = {
        Action.TYPE: parsed_args.keybinding_type,
        Action.COPY: parsed_args.keybinding_copy,
        Action.CLIPBOARD: parsed_args.keybinding_clipboard,
        Action.TYPE_NUMERICAL: parsed_args.keybinding_type_numerical,
        Action.UNICODE: parsed_args.keybinding_unicode,
        Action.COPY_UNICODE: parsed_args.keybinding_copy_unicode,
    }

    return parsed_args



from subprocess import run

from ..abstractionhelper import is_installed
from .typer import Typer


class CliclickTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_installed("cliclick")

    @staticmethod
    def name() -> str:
        return "cliclick"

    def get_active_window(self) -> str:
        return ""

    def type_characters(self, characters: str, active_window: str) -> None:
        pass

    def insert_from_clipboard(self, active_window: str) -> None:
        run(["cliclick", "-w 100", "kd:cmd", "t:v", "ku:cmd"])

    def type_numerical(self, codepoints: list[int], active_window: str) -> None:
        pass



import argparse
import json
import os
import re
import sys
from dataclasses import asdict, dataclass
from enum import IntEnum, auto

from . import emoji_data
from .action import execute_action
from .argument_parsing import parse_arguments_flexible
from .clipboarder.clipboarder import Clipboarder
from .emoji_data import fitzpatrick_modifiers, fitzpatrick_modifiers_reversed, skin_tone_selectable_emojis
from .file_loader import read_characters_from_files
from .frecent import load_frecent_characters, save_frecent_characters
from .models import Action, CharacterEntry
from .paths import cache_file_location
from .recent import load_recent_characters, save_recent_characters
from .typer.typer import Typer


class Step(IntEnum):
    SHOW_ALL = auto()
    SHORTCUTS = auto()
    SELECT_SKIN_TONE = auto()
    SELECT_ACTION = auto()
    EXECUTE = auto()
    DONE = auto()


@dataclass(slots=True)
class State:
    step: Step
    actions: list[Action]
    processed_characters: str
    unprocessed_characters: list[str]
    return_code: int
    __current_input: str | None = None
    output: str | None = None

    def __post_init__(self):
        if isinstance(self.step, int) and not isinstance(self.step, Step):
            self.step = Step(self.step)
        if self.actions and isinstance(self.actions[0], str):
            self.actions = [Action(a) for a in self.actions]

    def save_to_cache(self) -> None:
        with cache_file_location.open("w+") as file:
            json.dump(asdict(self), file, default=str)

    @staticmethod
    def load_from_cache(current_input: str, return_code: int) -> "State":
        if not cache_file_location.is_file():
            return State(
                step=Step.SHOW_ALL,
                actions=[],
                processed_characters="",
                unprocessed_characters=[],
                return_code=return_code,
                output=None,
            )

        with cache_file_location.open("r") as file:
            state = State(**json.load(file))
        state.__current_input = current_input
        state.return_code = return_code
        state.output = None
        return state

    @staticmethod
    def remove_cache():
        try:
            cache_file_location.unlink()
        except FileNotFoundError:
            pass

    @property
    def has_input(self) -> bool:
        return self.__current_input is not None

    @property
    def current_input(self) -> str:
        temp = self.__current_input
        self.__current_input = None
        return temp

    def reset_current_input(self) -> None:
        self.__current_input = None


class ModeRofimoji:
    args: argparse.Namespace
    typer: Typer
    clipboarder: Clipboarder

    def mode(self) -> None:
        if os.environ.get("ROFI_RETV") == "0":
            State.remove_cache()

        state = State.load_from_cache(os.environ.get("ROFI_INFO", sys.argv[-1]), int(os.environ.get("ROFI_RETV", "")))
        self.__parse_args()
        state.actions = self.args.actions

        if state.step == Step.SHOW_ALL:
            self.show_characters(state)
        if state.step == Step.SHORTCUTS:
            self.handle_shortcuts(state)
        if state.step == Step.SELECT_SKIN_TONE:
            self.select_skin_tone(state)
        if state.step == Step.SELECT_ACTION:
            self.choose_action(state)
        if state.step == Step.EXECUTE:
            self.execute_actions(state)
        if state.step == Step.DONE:
            state.remove_cache()
        else:
            state.save_to_cache()

        if state.output:
            print(state.output)

    def __parse_args(self) -> None:
        self.args = parse_arguments_flexible()
        self.typer = Typer.best_option(self.args.typer)
        self.clipboarder = Clipboarder.best_option(self.args.clipboarder)

    def show_characters(self, state: State) -> None:
        recent_characters = self.__format_recent_characters(
            load_recent_characters(self.args.max_recent, self.args.files)
        )

        state.output = "\x00markup-rows\x1ftrue\n"
        state.output += "\x00use-hot-keys\x1ftrue\n"
        if len(recent_characters) > 0:
            state.output += f"\x00message\x1f{recent_characters}\n"
        state.output += "\n".join(
            self.__format_characters(
                read_characters_from_files(
                    self.args.files, load_frecent_characters() if self.args.frecency else [], self.args.use_additional
                )
            )
        )
        state.output += "\n"

        state.step += 1

    def __format_recent_characters(self, recent_characters: list[str]) -> str:
        pairings = [f"\u200e{(index + 1) % 10}: {character}" for index, character in enumerate(recent_characters)]

        return " | ".join(pairings)

    def __format_characters(self, characters: list[CharacterEntry]) -> list[str]:
        if self.args.use_icons and not self.args.show_description:
            return [
                f" \0meta\x1f{entry.description}\x1ficon\x1f<span>{entry.character}</span>\x1finfo\x1f{entry.character}"
                for entry in characters
            ]
        elif self.args.use_icons and self.args.show_description:
            return [
                f"{entry.description}\0icon\x1f<span>{entry.character}</span>\x1finfo\x1f{entry.character}"
                for entry in characters
            ]
        elif not self.args.use_icons and self.args.show_description:
            return [f"{entry.character} {entry.description}" for entry in characters]
        else:
            return [f"{entry.character}\0meta\x1f{entry.description}" for entry in characters]

    def handle_shortcuts(self, state: State) -> None:
        if 10 <= state.return_code <= 19:
            state.processed_characters = load_recent_characters(self.args.max_recent, self.args.files)[
                state.return_code - 10
            ]
            state.reset_current_input()
            state.step += 2
            return
        elif state.return_code:
            new_actions = self.__choose_action_from_return_code(state.return_code)
            if new_actions:
                state.actions = new_actions
            state.unprocessed_characters = state.current_input.splitlines()
            state.step += 1
            return
        else:
            return

    def __choose_action_from_return_code(self, return_code: int) -> list[Action]:
        match return_code:
            case 20:
                return [Action.COPY]
            case 21:
                return [Action.TYPE]
            case 22:
                return [Action.CLIPBOARD]
            case 23:
                return [Action.TYPE_NUMERICAL]
            case 24:
                return [Action.UNICODE]
            case 25:
                return [Action.COPY_UNICODE]
            case _:
                return []

    def __extract_char_from_input(self, line) -> str:
        return re.match(r"^(?:\u200e(?! ))?(?P<char>.[^ ]*)( .*|$)", line).group("char")

    def select_skin_tone(self, state: State) -> None:
        if state.has_input:
            state.processed_characters += self.__extract_char_from_input(state.current_input)
            state.unprocessed_characters.pop(0)

        while state.unprocessed_characters:
            character = self.__extract_char_from_input(state.unprocessed_characters[0])
            save_frecent_characters(character)

            if character in emoji_data.skin_tone_selectable_emojis and self.args.skin_tone == "ask":
                if self.args.use_icons and not self.args.show_description:
                    state.output = "\n".join(
                        f" \0meta\x1f{description}\x1ficon\x1f<span>{character}{modifier}</span>\x1finfo\x1f{character}{modifier}"
                        for (modifier, description) in fitzpatrick_modifiers.items()
                    )
                elif self.args.use_icons and self.args.show_description:
                    state.output = "\n".join(
                        f"{description}\0icon\x1f<span>{character}{modifier}</span>\x1finfo\x1f{character}{modifier}"
                        for (modifier, description) in fitzpatrick_modifiers.items()
                    )
                elif not self.args.use_icons and self.args.show_description:
                    state.output = "\n".join(
                        f"{character}{modifier} {description}"
                        for (modifier, description) in fitzpatrick_modifiers.items()
                    )
                else:
                    state.output = "\n".join(
                        f"{character}{modifier}\0meta\x1f{description}"
                        for (modifier, description) in fitzpatrick_modifiers.items()
                    )
                return

            if character not in skin_tone_selectable_emojis or self.args.skin_tone == "neutral":
                state.processed_characters += character
            else:
                state.processed_characters += character + fitzpatrick_modifiers_reversed[self.args.skin_tone]
            state.unprocessed_characters.pop(0)

        state.step += 1

    def choose_action(self, state: State) -> None:
        if state.has_input:
            state.actions = [Action(state.current_input)]
            state.step += 1
            return

        if Action.MENU in state.actions:
            state.output = "\n".join([str(it) for it in Action if it != Action.MENU])
            return

        state.step += 1

    def execute_actions(self, state: State) -> str | None:
        save_recent_characters(state.processed_characters, self.args.max_recent, self.args.files)
        execute_action(state.processed_characters, state.actions, "")
        state.step += 1
        return



from subprocess import run

from ..abstractionhelper import is_installed
from ..models import CANCEL, DEFAULT, Action, CharacterEntry, Shortcut
from .selector import Selector


class Bemenu(Selector):
    @staticmethod
    def supported() -> bool:
        return is_installed("bemenu")

    @staticmethod
    def name() -> str:
        return "bemenu"

    def show_character_selection(
        self,
        characters: list[CharacterEntry],
        recent_characters: list[str],
        prompt: str,
        show_description: bool,
        use_icons: bool,
        keybindings: dict[Action, str],
        additional_args: list[str],
    ) -> tuple[Action | DEFAULT | CANCEL, list[str] | Shortcut]:
        parameters = ["bemenu", "--ignorecase", "-p", prompt, *additional_args]

        bemenu = run(
            parameters, input="\n".join(self.basic_format_characters(characters)), capture_output=True, encoding="utf-8"
        )
        return DEFAULT(), [self._extract_char_from_output(line) for line in bemenu.stdout.splitlines()]

    def show_skin_tone_selection(
        self,
        selected_emoji: str,
        prompt: str,
        show_description: bool,
        use_icons: bool,
        additional_args: list[str],
    ) -> tuple[int, str]:
        bemenu = run(
            ["bemenu", "-p", prompt, *additional_args],
            input="\n".join(self._basic_format_skin_tones(selected_emoji)),
            capture_output=True,
            encoding="utf-8",
        )

        return bemenu.returncode, self._extract_char_from_output(bemenu.stdout)

    def show_action_menu(self, additional_args: list[str]) -> list[Action]:
        bemenu = run(
            [
                "bemenu",
                *additional_args,
            ],
            input="\n".join([it.value for it in Action if it != Action.MENU]),
            capture_output=True,
            encoding="utf-8",
        )

        return [Action(bemenu.stdout.strip())]



from .clipboarder.clipboarder import Clipboarder
from .models import Action
from .typer.typer import Typer


def execute_action(
    characters: str,
    actions: list[Action],
    active_window: str,
    typer_preference: str | None = None,
    clipboarder_preference: str | None = None,
) -> None:
    typer = Typer.best_option(typer_preference)
    clipboarder = Clipboarder.best_option(clipboarder_preference)

    for action in actions:
        match action:
            case Action.TYPE:
                typer.type_characters(characters, active_window)
            case Action.COPY:
                clipboarder.copy_characters_to_clipboard(characters)
            case Action.CLIPBOARD:
                clipboarder.copy_paste_characters(characters, active_window, typer)
            case Action.TYPE_NUMERICAL:
                typer.type_numerical(__as_codepoints(characters), active_window)
            case Action.UNICODE:
                typer.type_characters(__as_codepoint_string(characters), active_window)
            case Action.COPY_UNICODE:
                clipboarder.copy_characters_to_clipboard(__as_codepoint_string(characters))
            case Action.STDOUT:
                print(characters)


def __as_codepoints(characters: str) -> list[int]:
    return [ord(c) for c in characters]


def __as_codepoint_string(characters: str) -> str:
    return "-".join(f"{c:x}" for c in __as_codepoints(characters))



import os
from glob import glob
from pathlib import Path

from .models import CharacterEntry
from .paths import custom_additional_files_location


def read_characters_from_files(files: list[str], frecent: list[str], use_additional: bool) -> list[CharacterEntry]:
    all_characters: dict[str, CharacterEntry] = {}

    for character in frecent:
        all_characters[character] = CharacterEntry(character)

    for file in __resolve_all_filenames(files, use_additional):
        characters_from_file = __load_from_file(file)
        for character_entry in characters_from_file:
            if character_entry.character in all_characters:
                all_characters[character_entry.character].merge(character_entry)
            else:
                all_characters[character_entry.character] = character_entry

    return [character for character in all_characters.values() if character.description]


def __resolve_all_filenames(file_names: list[str], use_additional: bool) -> list[Path]:
    resolved_file_names = []
    for file_name in file_names:
        resolved_file_names += __resolve_filename(file_name, use_additional)

    return resolved_file_names


def __resolve_filename(file_name: str, use_additional: bool) -> list[Path]:
    resolved_file_names = []

    for absolute_file in glob(os.path.expanduser(file_name)):
        resolved_file_names.append(Path(absolute_file))

    if resolved_file_names:
        return resolved_file_names

    for custom_files in custom_additional_files_location.glob(file_name if "*" in file_name else f"{file_name}*"):
        if not custom_files.name.endswith(".additional.csv"):
            resolved_file_names.append(custom_files)

    if resolved_file_names:
        return resolved_file_names

    for distributed_file in (Path(__file__).parent / "data").glob(file_name if "*" in file_name else f"{file_name}*"):
        resolved_file_names.append(distributed_file)
        resolved_file_names += __load_additional_files(distributed_file, use_additional)

    if resolved_file_names:
        return resolved_file_names

    if file_name == "all":
        nested_file_names = [
            __resolve_filename(file.stem, use_additional) for file in (Path(__file__).parent / "data").glob("*.csv")
        ]
        resolved_file_names += [file_name for file_names in nested_file_names for file_name in file_names]
        return resolved_file_names

    raise FileNotFoundError(f"Couldn't find file {file_name!r}")


def __load_additional_files(original_file: Path, use_additional: bool) -> list[Path]:
    additional_files = []
    custom_additional_file = custom_additional_files_location / f"{original_file.stem}.additional.csv"
    if custom_additional_file.is_file():
        additional_files.append(custom_additional_file)
    provided_additional_file = Path(__file__).parent / "data" / "additional" / f"{original_file.stem}.csv"
    if use_additional and provided_additional_file.is_file():
        additional_files.append(provided_additional_file)

    return additional_files


def __load_from_file(file: Path) -> list[CharacterEntry]:
    lines = file.read_text().strip("\n").split("\n")

    all_character_entries = []
    for line in lines:
        parsed_line = line.split(" ", 1)
        all_character_entries.append(CharacterEntry(parsed_line[0], parsed_line[1]))

    return all_character_entries



from subprocess import run

from ..abstractionhelper import is_installed
from .typer import Typer


class YdotoolTyper(Typer):
    @staticmethod
    def name():
        return "ydotool"

    @staticmethod
    def supported():
        return is_installed("ydotool")

    def get_active_window(self):
        return "not possible with ydotool"

    def type_characters(self, characters: str, active_window: str) -> None:
        pass

    def type_numerical(self, codepoints: list[int], active_window: str) -> None:
        keypresses = []
        for codepoint in codepoints:
            keypresses.append = [
                (self.__get_event_code("LeftCtrl") + ":1"),
                (self.__get_event_code("LeftShift") + ":1"),
                (self.__get_event_code("U") + ":1"),
                (self.__get_event_code("U") + ":0"),
                (self.__get_event_code("LeftShift") + ":0"),
                (self.__get_event_code("LeftCtrl") + ":0"),
                (self.__get_event_code(f"{codepoint:x}") + ":1"),
                (self.__get_event_code(f"{codepoint:x}") + ":0"),
            ]

        run(["ydotool", "key", *keypresses])

    def insert_from_clipboard(self, active_window: str) -> None:
        run(
            [
                "ydotool",
                "key",
                (self.__get_event_code("LeftShift") + ":1"),
                (self.__get_event_code("Insert") + ":1"),
                (self.__get_event_code("Insert") + ":0"),
                (self.__get_event_code("LeftShift") + ":0"),
            ]
        )

    def __get_event_code(self, char: str) -> str:
        return str(self._keycodes[char])

    # From /usr/include/linux/input-event-codes.h
    _keycodes = {
        "0": 11,
        "1": 2,
        "2": 3,
        "3": 4,
        "4": 5,
        "5": 6,
        "6": 7,
        "7": 8,
        "8": 9,
        "9": 10,
        "a": 30,
        "b": 48,
        "c": 46,
        "d": 32,
        "e": 18,
        "f": 33,
        "u": 22,
        "Enter": 28,
        "LeftCtrl": 29,
        "LeftShift": 42,
        "BackSpace": 14,
        "Tab": 15,
        "Insert": 110,
    }



from subprocess import run

from ..abstractionhelper import is_installed, is_wayland
from .typer import Typer


class WTypeTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("wtype")

    @staticmethod
    def name() -> str:
        return "wtype"

    def get_active_window(self) -> str:
        return "not possible with wtype"

    def type_characters(self, characters: str, active_window: str) -> None:
        run(["wtype", characters])

    def insert_from_clipboard(self, active_window: str) -> None:
        run(["wtype", "-M", "shift", "-P", "Insert", "-p", "Insert", "-m", "shift"])

    def type_numerical(self, codepoints: list[int], active_window: str) -> None:
        codepoint_list = " ".join([f"U{codepoint:x}" for codepoint in codepoints])

        run(["wtype", "-M", "ctrl", "-M", "shift", codepoint_list, "-m", "ctrl", "-m", "shift"])



import hashlib
from pathlib import Path

from .paths import recents_file_location


def load_recent_characters(max_recent: int, files: list[str]) -> list[str]:
    try:
        return [char.strip("\n") for char in __filename_for(files).read_text().strip("\n").split("\n")][:max_recent]
    except FileNotFoundError:
        return []
    except NotADirectoryError:
        recents = [char.strip("\n") for char in recents_file_location.read_text().strip("\n").split("\n")][:max_recent]
        recents_file_location.unlink()
        save_recent_characters("".join(recents), max_recent, files)
        return recents


def save_recent_characters(new_characters: str, max_recent: int, files: list[str]) -> None:
    if max_recent == 0:
        return
    max_recent = min(max_recent, 10)

    old_file_name = __filename_for(files)
    new_file_name = old_file_name.with_suffix(".tmp")

    new_file_name.parent.mkdir(parents=True, exist_ok=True)
    with new_file_name.open("w+") as new_file:
        new_file.write(new_characters + "\n")

        try:
            with old_file_name.open("r") as old_file:
                index = 0
                for line in old_file:
                    if new_characters != line.strip("\n"):
                        if index == max_recent - 1:
                            break
                        new_file.write(line)
                        index += 1

            old_file_name.unlink()
        except FileNotFoundError:
            pass

    new_file_name.rename(old_file_name)


def __filename_for(files: list[str]) -> Path:
    return recents_file_location / hashlib.sha256(b"\0".join(f.encode() for f in sorted(files))).hexdigest()



from abc import ABC, abstractmethod


class Typer(ABC):
    @staticmethod
    def best_option(name: str | None = None) -> "Typer":
        from .cliclick import CliclickTyper
        from .noop import NoopTyper
        from .wl_ime_type import WlImeTypeTyper
        from .wtype import WTypeTyper
        from .xdotool import XDoToolTyper
        from .ydotool import YdotoolTyper as YDoToolTyper

        available_typers = [XDoToolTyper, WTypeTyper, YDoToolTyper, CliclickTyper, WlImeTypeTyper, NoopTyper]

        if name is not None:
            return next(typer for typer in available_typers if typer.name() == name)()
        else:
            return next(typer for typer in available_typers if typer.supported())()

    @staticmethod
    @abstractmethod
    def supported() -> bool:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @abstractmethod
    def get_active_window(self) -> str:
        pass

    @abstractmethod
    def type_characters(self, characters: str, active_window: str) -> None:
        pass

    @abstractmethod
    def insert_from_clipboard(self, active_window: str) -> None:
        pass

    @abstractmethod
    def type_numerical(self, codepoints: list[int], active_window: str) -> None:
        pass



import math

from .paths import frecency_file_location


def load_frecent_characters() -> list[str]:
    return list(__load_frecent_characters().keys())


def __load_frecent_characters() -> dict[str, float]:
    frecencies = {}
    try:
        with frecency_file_location.open("r") as file:
            for line in file:
                (frecency, character) = line.strip("\n").split(" ")
                frecencies[character] = float(frecency)
    except FileNotFoundError:
        pass

    return frecencies


def save_frecent_characters(chosen_character: str) -> None:
    new_file_name = frecency_file_location.with_name("frecency.tmp")

    new_file_name.parent.mkdir(parents=True, exist_ok=True)

    frecencies = __load_frecent_characters()

    frecencies[chosen_character] = frecencies.get(chosen_character, 0) + 1.1

    with new_file_name.open("w+") as new_file:
        for character, frecency in sorted(frecencies.items(), key=lambda item: item[1], reverse=True):
            new_file.write(f"{math.floor(frecency)} {character}\n")

    new_file_name.rename(frecency_file_location)



from subprocess import run

from ..abstractionhelper import is_installed
from ..typer.typer import Typer
from .clipboarder import Clipboarder


class PBCopyClipboarder(Clipboarder):
    @staticmethod
    def supported() -> bool:
        return is_installed("pbcopy")

    @staticmethod
    def name() -> str:
        return "pbcopy"

    def copy_characters_to_clipboard(self, characters: str) -> None:
        run(["pbcopy"], input=characters, encoding="utf-8")

    def copy_paste_characters(self, characters: str, active_window: str, typer: Typer) -> None:
        old_clipboard_content = run(args=["pbpaste"], capture_output=True, encoding="utf-8")

        run(args=["pbcopy"], input=characters, encoding="utf-8")

        typer.insert_from_clipboard(active_window)

        if old_clipboard_content.returncode == 0:
            run(args=["pbcopy"], input=old_clipboard_content.stdout, encoding="utf-8")



import subprocess
from subprocess import run

from ..abstractionhelper import is_installed, is_wayland
from ..typer.typer import Typer
from .clipboarder import Clipboarder


class XClipClipboarder(Clipboarder):
    @staticmethod
    def supported() -> bool:
        return not is_wayland() and is_installed("xclip")

    @staticmethod
    def name() -> str:
        return "xclip"

    def copy_characters_to_clipboard(self, characters: str) -> None:
        run(
            ["xclip", "-i", "-selection", "clipboard"],
            input=characters,
            encoding="utf-8",
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )

    def copy_paste_characters(self, characters: str, active_window: str, typer: Typer) -> None:
        old_clipboard_content = run(
            args=["xclip", "-out", "-selection", "clipboard", "-target", "text/plain;charset=utf-8"],
            capture_output=True,
            encoding="utf-8",
        )
        old_primary_content = run(
            args=["xclip", "-out", "-selection", "primary", "-target", "text/plain;charset=utf-8"],
            capture_output=True,
            encoding="utf-8",
        )

        run(
            args=["xclip", "-i", "-selection", "clipboard"],
            input=characters,
            encoding="utf-8",
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )
        run(
            args=["xclip", "-i", "-selection", "primary"],
            input=characters,
            encoding="utf-8",
            stderr=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
        )

        typer.insert_from_clipboard(active_window)

        if old_clipboard_content.returncode == 0:
            run(
                args=["xclip", "-i", "-selection", "clipboard"],
                input=old_clipboard_content.stdout,
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                encoding="utf-8",
            )
        if old_primary_content.returncode == 0:
            run(
                args=["xclip", "-i", "-selection", "primary"],
                input=old_primary_content.stdout,
                stderr=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                encoding="utf-8",
            )



from .typer import Typer


class NoopTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return True

    @staticmethod
    def name() -> str:
        return "noop"

    def get_active_window(self) -> str:
        return ""

    def type_characters(self, characters: str, active_window: str) -> None:
        raise NoTyperFoundException()

    def insert_from_clipboard(self, active_window: str) -> None:
        raise NoTyperFoundException()

    def type_numerical(self, codepoints: list[int], active_window: str) -> None:
        raise NoTyperFoundException()


class NoTyperFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to type characters. Please check the required dependencies."



import sys

from . import emoji_data
from .action import execute_action
from .argument_parsing import parse_arguments_strict
from .clipboarder.clipboarder import Clipboarder
from .file_loader import read_characters_from_files
from .frecent import load_frecent_characters, save_frecent_characters
from .models import CANCEL, DEFAULT, Action, Shortcut
from .recent import load_recent_characters, save_recent_characters
from .selector.selector import Selector
from .typer.typer import Typer


class StandaloneRofimoji:
    def __init__(self) -> None:
        self.args = parse_arguments_strict()
        self.selector = Selector.best_option(self.args.selector)
        self.typer = Typer.best_option(self.args.typer)
        self.clipboarder = Clipboarder.best_option(self.args.clipboarder)
        self.active_window = self.typer.get_active_window()

    def standalone(self) -> None:
        action, value = self.__open_main_selector_window()

        if action == CANCEL():
            sys.exit(1)
        elif action != DEFAULT():
            self.args.actions = [action]

        if isinstance(value, Shortcut):
            characters = load_recent_characters(self.args.max_recent, self.args.files)[value.index]
        else:
            characters = self.__process_chosen_characters(value)

        if Action.MENU in self.args.actions:
            self.args.actions = self.selector.show_action_menu(self.args.selector_args)

        save_recent_characters(characters, self.args.max_recent, self.args.files)
        execute_action(characters, self.args.actions, self.active_window, self.args.typer, self.args.clipboarder)

    def __open_main_selector_window(self) -> tuple[Action | DEFAULT | CANCEL, list[str] | Shortcut]:
        return self.selector.show_character_selection(
            read_characters_from_files(
                self.args.files, load_frecent_characters() if self.args.frecency else [], self.args.use_additional
            ),
            load_recent_characters(self.args.max_recent, self.args.files),
            self.args.prompt,
            self.args.show_description,
            self.args.use_icons,
            self.args.keybindings,
            self.args.selector_args,
        )

    def __process_chosen_characters(self, characters: list[str]) -> str:
        characters_with_skin_tones = []
        for character in characters:
            save_frecent_characters(character)
            characters_with_skin_tones.append(self.__add_skin_tone_to_character(character))

        return "".join(characters_with_skin_tones)

    def __add_skin_tone_to_character(self, character: str) -> str:
        characters_with_skin_tone = []

        for element in character:
            if element in emoji_data.skin_tone_selectable_emojis:
                characters_with_skin_tone.append(self.__select_skin_tone(element))
            else:
                characters_with_skin_tone.append(element)

        return "".join(characters_with_skin_tone)

    def __select_skin_tone(self, selected_emoji: str) -> str:
        skin_tone = self.args.skin_tone

        if skin_tone == "neutral":
            return selected_emoji
        elif skin_tone != "ask":
            return selected_emoji + emoji_data.fitzpatrick_modifiers_reversed[skin_tone]
        else:
            return_code, skin_tone = self.selector.show_skin_tone_selection(
                selected_emoji,
                selected_emoji + "   ",
                self.args.show_description,
                self.args.use_icons,
                self.args.selector_args,
            )

            if return_code == 1:
                return ""

            return skin_tone
